@@ -15,30 +15,29 @@
 #    along with Paperwork.  If not, see <http://www.gnu.org/licenses/>.
 
 import logging
-import multiprocessing
-import re
-import threading
 import time
 
 from gi.repository import GLib
 from gi.repository import GObject
+from gi.repository import Libinsane
 import pyocr
 import pyocr.builders
 
-from paperwork_backend.util import check_spelling
 from paperwork.frontend.mainwindow.pages import SimplePageDrawer
-from paperwork.frontend.util.jobs import Job
-from paperwork.frontend.util.jobs import JobFactory
 from paperwork.frontend.util.canvas.animations import Animation
 from paperwork.frontend.util.canvas.animations import ScanAnimation
 from paperwork.frontend.util.canvas.animations import SpinnerAnimation
-from paperwork.frontend.util.canvas.animators import LinearSimpleAnimator
 from paperwork.frontend.util.canvas.animators import LinearCoordAnimator
+from paperwork.frontend.util.canvas.animators import LinearSimpleAnimator
 from paperwork.frontend.util.canvas.drawers import fit
 from paperwork.frontend.util.canvas.drawers import LineDrawer
 from paperwork.frontend.util.canvas.drawers import PillowImageDrawer
 from paperwork.frontend.util.canvas.drawers import RectangleDrawer
 from paperwork.frontend.util.canvas.drawers import TargetAreaDrawer
+from paperwork.frontend.util.img import raw2pillow
+from paperwork.frontend.util.img import raw2pixbuf
+from paperwork.frontend.util.jobs import Job
+from paperwork.frontend.util.jobs import JobFactory
 
 
 logger = logging.getLogger(__name__)
@@ -57,7 +56,7 @@ class JobScan(Job):
         'scan-chunk': (GObject.SignalFlags.RUN_LAST, None,
                        # line where to put the image
                        (GObject.TYPE_INT,
-                        GObject.TYPE_PYOBJECT,)),  # The PIL image
+                        GObject.TYPE_PYOBJECT,)),  # Pixbuf
         'scan-done': (GObject.SignalFlags.RUN_LAST, None,
                       (GObject.TYPE_PYOBJECT, )),  # Pillow image
         'scan-error': (GObject.SignalFlags.RUN_LAST, None,
@@ -80,38 +79,57 @@ class JobScan(Job):
         self.emit('scan-started')
 
         try:
-            size = self.scan_session.scan.expected_size
-            self.emit('scan-info', size[0], size[1])
+            params = self.scan_session.get_scan_parameters()
+            self.emit('scan-info', params.get_width(), params.get_height())
+
+            assert(params.get_format() == Libinsane.ImgFormat.RAW_RGB_24)
+            line_length = params.get_width() * 3
 
             last_line = 0
-            try:
-                while self.can_run:
-                    self.scan_session.scan.read()
+            whole_image = []
+            chunk = bytearray()
 
-                    next_line = self.scan_session.scan.available_lines[1]
-                    if (next_line > last_line + 50):
-                        chunk = self.scan_session.scan.get_image(
-                            last_line, next_line
-                        )
-                        self.emit('scan-chunk', last_line, chunk)
-                        last_line = next_line
+            while self.can_run and not self.scan_session.end_of_page():
+                r = 0
+                while (r <= 512 * 1024 and
+                        not self.scan_session.end_of_page()):
+                    out = self.scan_session.read_bytes(64 * 1024)
+                    out = out.get_data()
+                    r += len(out)
+                    chunk.extend(out)
 
-                    time.sleep(0)  # Give some CPU time to Gtk
-                if not self.can_run:
-                    logger.info("Scan canceled")
-                    self.emit('scan-canceled')
-                    return
-            except EOFError:
-                pass
+                split = len(chunk) - (len(chunk) % line_length)
+                next_chunk = chunk[split:]
+                chunk = chunk[:split]
+
+                whole_image.append(chunk)
+
+                pixbuf = raw2pixbuf(chunk, params)
+                if pixbuf is None:
+                    continue
+
+                self.emit('scan-chunk', last_line, pixbuf)
+
+                time.sleep(0)  # Give some CPU time to Gtk
+                last_line += (split / line_length)
+                chunk = next_chunk
+
+            if not self.scan_session.end_of_feed():
+                self.scan_session.cancel()
+
+            if not self.can_run:
+                logger.info("Scan canceled")
+                self.emit('scan-canceled')
+                return
+
+            whole_image = b"".join(whole_image)
+            image = raw2pillow(whole_image, params)
+            self.emit('scan-done', image)
+            logger.info("Scan done")
         except Exception as exc:
             self.emit('scan-error', exc)
             raise
 
-        img = self.scan_session.images[-1]
-        self.emit('scan-done', img)
-        logger.info("Scan done")
-        del self.scan_session
-
     def stop(self, will_resume=False):
         self.can_run = False
         self._stop_wait()
@@ -152,79 +170,6 @@ class JobFactoryScan(JobFactory):
         return job
 
 
-class _ImgOCRThread(threading.Thread):
-    # we don't use jobs here, because we would need 1 scheduler for each job
-    # --> too painful and useless
-
-    def __init__(self, name, ocr_tool, langs, angle, img):
-        threading.Thread.__init__(self, name="OCR")
-        self.name = name
-        self.ocr_tool = ocr_tool
-        self.langs = langs
-        self.angle = angle
-        self.img = img
-        self.score = -1
-        self.boxes = None
-
-    def __compute_ocr_score_with_spell_checking(self, txt):
-        return check_spelling(self.langs['spelling'], txt)
-
-    @staticmethod
-    def __boxes_to_txt(boxes):
-        txt = u""
-        for line in boxes:
-            txt += line.content + u"\n"
-        return txt
-
-    @staticmethod
-    def __compute_ocr_score_without_spell_checking(txt):
-        """
-        Try to evaluate how well the OCR worked.
-        Current implementation:
-            The score is the number of words only made of 4 or more letters
-            ([a-zA-Z])
-        """
-        # TODO(Jflesch): i18n / l10n
-        score = 0
-        prog = re.compile(r'^[a-zA-Z]{4,}$')
-        for word in txt.split(" "):
-            if prog.match(word):
-                score += 1
-        return (txt, score)
-
-    def run(self):
-        SCORE_METHODS = [
-            ("spell_checker", self.__compute_ocr_score_with_spell_checking),
-            ("lucky_guess", self.__compute_ocr_score_without_spell_checking),
-            ("no_score", lambda txt: (txt, 0))
-        ]
-
-        logger.info("Running OCR on page orientation %s" % self.name)
-        self.boxes = self.ocr_tool.image_to_string(
-            self.img, lang=self.langs['ocr'],
-            builder=pyocr.builders.LineBoxBuilder())
-
-        txt = self.__boxes_to_txt(self.boxes)
-
-        for score_method in SCORE_METHODS:
-            try:
-                logger.info("Evaluating score of page orientation (%s)"
-                            " using method '%s' ..."
-                            % (self.name, score_method[0]))
-                (_, self.score) = score_method[1](txt)
-                # TODO(Jflesch): For now, we throw away the fixed version of
-                # the text:
-                # The original version may contain proper nouns, and spell
-                # checking could make them disappear
-                # However, it would be best if we could keep both versions
-                # without increasing too much indexation time
-                return
-            except Exception as exc:
-                logger.error("Scoring method '%s' on orientation %s failed !"
-                             % (score_method[0], self.name))
-                logger.error("Reason: %s" % exc)
-
-
 class JobOCR(Job):
     __gsignals__ = {
         'ocr-started': (GObject.SignalFlags.RUN_LAST, None,
@@ -262,8 +207,13 @@ class JobOCR(Job):
         if len(self.angles) == 1:
             orientation = {'angle': self.angles[0]}
         else:
-            orientation = self.ocr_tool.detect_orientation(
-                img, lang=self.langs['ocr'])
+            try:
+                orientation = self.ocr_tool.detect_orientation(
+                    img, lang=self.langs['ocr'])
+            except Exception as exc:
+                logger.error("OCR orientation detection failed", exc_info=exc)
+                # defaulting to default orientation
+                orientation = {'angle': self.angles[0]}
 
         if orientation['angle'] not in self.angles:
             raise Exception("OCR tool returned an unexpected orientation: %d"
@@ -291,61 +241,16 @@ class JobOCR(Job):
 
         return (orientation['angle'], img, boxes)
 
-    def do_ocr_with_custom_heuristic(self, img):
-        imgs = {angle: img.rotate(angle, expand=True) for angle in self.angles}
-        self.emit('ocr-angles', list(imgs.keys()))
-
-        if len(imgs) <= 0:
-            self.emit('ocr-score', 0, 0)
-            return (0, img, [])
-
-        max_threads = multiprocessing.cpu_count()
-        threads = []
-        scores = []
-
-        if len(imgs) > 1:
-            logger.debug("Will use %d process(es) for OCR" % (max_threads))
-
-        # Run the OCR tools in as many threads as there are processors/core
-        # on the computer
-        nb = 0
-        while (len(imgs) > 0 or len(threads) > 0):
-            # look for finished threads
-            for thread in threads:
-                if not thread.is_alive():
-                    threads.remove(thread)
-                    logger.info("OCR done on angle %d: %f"
-                                % (thread.angle, thread.score))
-                    scores.append((thread.score, thread.angle,
-                                   thread.img, thread.boxes))
-                    self.emit('ocr-score', thread.angle, thread.score)
-            # start new threads if required
-            while (len(threads) < max_threads and len(imgs) > 0):
-                (angle, img) = imgs.popitem()
-                logger.info("Starting OCR on angle %d" % angle)
-                thread = _ImgOCRThread(str(nb), self.ocr_tool,
-                                       self.langs, angle, img)
-                thread.start()
-                threads.append(thread)
-                nb += 1
-            time.sleep(self.OCR_THREADS_POLLING_TIME)
-
-        # We want the higher score first
-        scores.sort(key=lambda x: x[0])
-
-        logger.info("Best: %f" % (scores[0][0]))
-        return scores[0][1:]
-
     def do(self):
         self.emit('ocr-started', self.img)
 
         try:
             best = self.do_ocr_with_tool_heuristic(self.img)
         except Exception as exc:
-            logger.info("Failed to use OCR tool heuristic for orientation"
-                        " detection: %s" % str(exc))
-            logger.info("Falling back on Paperwork's heuristic")
-            best = self.do_ocr_with_custom_heuristic(self.img)
+            logger.error(
+                "Failed to use OCR tool heuristic for orientation"
+                " detection", exc_info=exc
+            )
 
         self.emit('ocr-done', best[0], best[1], best[2])
 
